#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/ip.h>
#include <linux/rtnetlink.h>
#include <net/rtnetlink.h>
#include <net/cfg80211.h>
#include <net/mac80211.h>
#include <linux/u64_stats_sync.h>

#define CRYPTO_ICMP_PROTO 	0x01
#define CRYPTO_IP_PROTO 	0x0800

static const char slave_name[]="wlan0";

static struct net_device_ops *orig_dataif_ops;
static struct net_device_ops new_dataif_ops;

static netdev_tx_t (*orig_start_xmit) (struct sk_buff *skb, struct net_device *dev);

inline void printk20b(unsigned char *data)
{
    
    printk("%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X \n",
    *(data), *(data+1), *(data+2), *(data+3), *(data+4), *(data+5), *(data+6), *(data+7), *(data+8), *(data+9), *(data+10),
    *(data+11), *(data+12), *(data+13), *(data+14), *(data+15), *(data+16), *(data+17), *(data+18), *(data+19) );
}

inline void pkt_encrypt(char * data, int len)
{
    int i;
    for (i = 0; i< len; i++)
    {
	*(data+i) = *(data+i) ^ 0x33;
    }
}

netdev_tx_t my_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
    if (CRYPTO_IP_PROTO == htons(skb->protocol))
    {
	if ( CRYPTO_ICMP_PROTO == ((struct iphdr *)(skb_network_header(skb)))->protocol)
	{
	    printk("TX icmp len = %d\n", htons(((struct iphdr *)(skb_network_header(skb)))->tot_len));
	    pkt_encrypt(  ((char*)(skb_network_header(skb)))  + 28 , htons(((struct iphdr *)(skb_network_header(skb)))->tot_len) - 28);
	}
    }
    return orig_start_xmit(skb, dev);
}

void show_ops(const struct net_device_ops *ops)
{
    printk("ops->ndo_open = %p\n", ops->ndo_open);
    printk("ops->ndo_stop = %p\n", ops->ndo_stop);
    printk("ops->ndo_uninit = %p\n", ops->ndo_uninit);
    printk("ops->ndo_start_xmit = %p\n", ops->ndo_start_xmit);
    printk("ops->ndo_set_rx_mode = %p\n", ops->ndo_set_rx_mode);
    printk("ops->ndo_change_mtu = %p\n", ops->ndo_change_mtu);
    printk("ops->ndo_set_mac_address = %p\n", ops->ndo_set_mac_address);
    printk("ops->ndo_select_queue = %p\n", ops->ndo_select_queue);
}

rx_handler_result_t handle_frame(struct sk_buff **pskb)
{
    struct sk_buff *skb = *pskb;
//  printk("RX len %d datalen %d maclen %d hdrlen %d proto 0x%4X truesize %d\n",(skb->len),htons(skb->data_len),htons(skb->mac_len),htons(skb->hdr_len),htons(skb->protocol),(skb->truesize));
    if (CRYPTO_IP_PROTO == htons(skb->protocol))
    {
	if ( CRYPTO_ICMP_PROTO == ((struct iphdr *)(skb_network_header(skb)))->protocol ) 
	{
	    printk("RX icmp len = %d\n", htons(((struct iphdr *)(skb_network_header(skb)))->tot_len));       
	
	    pkt_encrypt(  ((char*)(skb_network_header(skb)))  + 28 , htons(((struct iphdr *)(skb_network_header(skb)))->tot_len) - 28);
	
	}   
    }

    return RX_HANDLER_PASS;
}



int __init crypto_init(void)
{
    struct net_device *slave;

    rtnl_lock();

    slave = __dev_get_by_name(&init_net, slave_name);
    if (!slave)
        return -ENODEV;


    printk("%s->netdev_ops=%p\n", slave_name, slave->netdev_ops);

    if (slave->netdev_ops)
    {
        printk("%s->netdev_ops->ndo_start_xmit =%p\n", slave_name, slave->netdev_ops->ndo_start_xmit);

        /* save original net_device options*/
        orig_dataif_ops = slave->netdev_ops;
        new_dataif_ops = *slave->netdev_ops;

        /* update ndo_start_xmit */
        orig_start_xmit = slave->netdev_ops->ndo_start_xmit;

        printk("Original options:\n");
        show_ops(slave->netdev_ops);

        printk("Stored options:\n");
        show_ops(&new_dataif_ops);

        /* change net_device to new ones */
        new_dataif_ops.ndo_start_xmit = my_start_xmit;

        printk("New options:\n");
        show_ops(&new_dataif_ops);

        slave->netdev_ops = &new_dataif_ops;
    }
    netdev_rx_handler_register(slave,handle_frame,NULL);
    printk("loaded\n");

    netdev_priv(slave);

    rtnl_unlock();

    return 0;
}

void __exit crypto_exit(void)
{
    struct net_device *slave;

    rtnl_lock();

    slave = __dev_get_by_name(&init_net, slave_name);
    if (slave)
    {
        printk("%s->netdev_ops=%p\n", slave_name, slave->netdev_ops);
        netdev_rx_handler_unregister(slave);
        //wiphy_priv(slave->ieee80211_ptr->wiphy)
        if (slave->netdev_ops)
        {
             printk("%s->netdev_ops->ndo_start_xmit =%p\n", slave_name, slave->netdev_ops->ndo_start_xmit);

             /* restire original net_device options*/
             slave->netdev_ops = orig_dataif_ops;

             printk("Restored options:\n");
             show_ops(slave->netdev_ops);
        }
    }

    rtnl_unlock();

    printk("unloaded\n");
}

module_init(crypto_init);
module_exit(crypto_exit);
MODULE_LICENSE("GPL");
